/**
   This code is released under the terms of the MIT license
   (https://opensource.org/licenses/MIT).

   Copyright 2020 Stephan Beal (https://wanderinghorse.net)

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation
   files (the "Software"), to deal in the Software without
   restriction, including without limitation the rights to use, copy,
   modify, merge, publish, distribute, sublicense, and/or sell copies
   of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

/**
   The basic app-level code for Dark Venture Card Tools apps, as
   distinct from the library-level code (see dv.d/dv.js).
*/
(function(global){
    "use strict";
    const DV = global.DarkVenture, D = DV.dom;

    /**
       Internal helper which tries to reduce its argument to a string,
       accounting for various types of inputs.
    */
    const messageArgToString = function f(a){
        //console.debug("messageArgToString()",a);
        if(!a
           || 'string'===typeof a
           || 'number'===typeof a
           || 'boolean'===typeof a){
            return ''+a;
        }
        else if(a instanceof Error || a.message){
            return f(a.message);
        }else if(a.exception){/* From certain AJAX APIs */
            return f(a.exception.message);
        }else if(a.errorJson && a.errorJson.exception){
            /* (Historical) From whajaj APIs. */
            return a.errorJson.exception.message;
        }else if(a.errorMessage){
            /* (Historical) From whajaj APIs. */
            return a.errorMessage;
        }
        return a;
    };
    const argsToArray = (a)=>Array.prototype.slice.call(a,0);

    const App = DV.App = {

        config: {},

        /**
           App objects, for purposes of creating a link of pages
           (apps).
        */
        theApps: [{
            /** Unique ID */
            id: 'sourcerepo',
            /** Menu entry label */
            name: 'Project Home',
            /** Page title and main header text (quest-app pages only). */
            title: '',
            httpOnly: true, /* or fileOnly, for use when browsing locally */
            /** <A href=...> */
            href: window.location.pathname.replace(/\/doc\/.+$/,'/')
        },{
            id: 'sourcerepo',
            name: 'Project Home',
            title: 'irrelevant',
            fileOnly: true,
            href: 'https://fossil.wanderinghorse.net/r/darkventure'
        },{
            id: 'index',
            name: 'Main Docs',
            title: 'Dark Venture Card Tools',
            href: 'index.html',
            /**
               Colors are expanded to dv-fg-XXX and dv-bg-XXX CSS classes, so must
               refer to such color names.
            */
            headerBg: 'red',
            headerFg: 'white',
            cardBorders: true
        },{
            id: 'sidequests',
            name: 'Side Quests',
            title: 'Dark Venture Fan-Created Side Quests',
            href: 'sidequests.html',
            headerBg: 'lightblue',
            headerFg: 'black',
            toolbar: true,
            addStatusMessageArea: false
        },{
            id: 'heroicquests',
            name: 'Heroic Quests',
            title: 'Dark Venture Fan-created Heroic Quests',
            href: 'heroicquests.html',
            headerBg: 'orange',
            headerFg: 'black',
            toolbar: true,
            addStatusMessageArea: false
        },{
            id: 'scratchpad',
            name: 'Scratchpad',
            title: 'Dark Venture Card Scratchpad',
            href: 'scratchpad.html',
            noMenuEntry: true,
            addStatusMessageArea: true
        }],

        /** Cached DOM elements. */
        e: {
            appHeader: document.querySelector('#dv-app-header')
        },

        appendToHeader: function(e){
            const last = this.e.toolbar || this.e.appHeader.lastElementChild;
            this.e.appHeader.insertBefore(e, last);
            return this;
        },
        
        /** Reveals or hides the status message bar. */
        _showStatusBar: function(show){
            if(show){
                D.removeClass(this.e.statusMessageAreaHolder, 'hidden');
                D.addClass(this.e.appHeader, 'has-status-message-area');
            }else{
                D.addClass(this.e.statusMessageAreaHolder, 'hidden');
                D.removeClass(this.e.appHeader, 'has-status-message-area');
            }
            return this;
        },
        
        /**
           Outputs its arguments to the app's status message area,
           taking care to convert certain expected types of arguments
           to strings.
        */
        message: function(){
            const e = this.e.statusMessageArea,
                  ep = this.e.statusMessageAreaHolder;
            if(!e || !ep) return this;
            D.removeClass(ep, 'achtung');
            const args = arguments.length
                  ? argsToArray(arguments).map((a)=>{
                      if(a instanceof Error){
                          D.addClass(e, 'achtung');
                      }
                      return messageArgToString(a);
                  }) : [];
            e.innerText = args.join(' ');
            if(args.length){
                this._showStatusBar(true);
                args.unshift('DV.App:');
                console.info.apply(console, args);
            }else{
                this._showStatusBar(false);
            }
            return this;
        },

        /** The "error" counterpart of this.message(). */
        error: function(){
            const argv = argsToArray(arguments);
            console.error.apply(console, argv);
            const e = this.e.statusMessageArea,
                  ep = this.e.statusMessageAreaHolder;
            if(!e || !ep) return this;
            if(!D.isElementKindaInViewport(ep,1)){
                ep.scrollIntoView();
                // ^^ if we unconditionally do that, it's always scrolling that to
                // the very top of the viewport, which is irritating.
            }
            D.addClass(ep, 'achtung');
            const args = argv.map(messageArgToString);
            e.innerText = args.join(' ');
            if(args.length){
                this._showStatusBar(true);
                args.unshift('DV.App:');
                console.error.apply(console, args);
            }
            else{
                this._showStatusBar(false);
            }
            return this;
        },

        /** Temporary internal helper - gets removed after use. */
        _addStatusMessageArea: function(){
            const h = this.e.appHeader;
            const holder = D.attr(
                D.addClass(D.div(),'hidden','dv-no-print'),
                'id',
                'status-message-area-holder'
            );
            D.addClass(h, 'has-status-message-area');
            const sma = D.attr(
                    D.append(
                        D.div(),
                        "status messages will go here"
                    ),
                'id',
                'status-message-area'
            );
            D.append(holder, sma)
            h.parentNode.insertBefore(holder, h.nextSibling);
            //D.append(h,holder);
            const clear = ()=>this.message();
            const btnClear = D.button("Clear");
            holder.append(btnClear);
            sma.addEventListener('dblclick', clear, false);
            btnClear.addEventListener('click', clear, false);
            this.e.statusMessageArea = sma;
            this.e.statusMessageAreaHolder = holder;
        },

        /**
           Sets up element #dv-app-header, if found, to inject a
           header and a list of links to known apps/pages. If appId
           refers to one of this.theApps then the menu entry for that
           entry is styled differently to indicate it's the current
           page. If the referenced app object's toolbar property is
           true, a toolbar of card-related options is also installed.
           Returns this object.
        */
        setupAppHeader: function(appId){
            const body = document.body,
                  header = this.e.appHeader;
            if(!header) return this;
            if(appId) D.addClass(body,"dv-app-id-"+appId);
            const menu = D.attr(D.div(),'id','dv-app-menu');
            const h1 = D.h(1);
            var thisApp;
            const self = this;
            const isHttp = /^http/.test(window.location.protocol);
            this.theApps.forEach(function(a){
                if(a.httpOnly && !isHttp) return;
                else if(a.fileOnly && isHttp) return;
                else if(appId === a.id){
                    thisApp = a;
                    D.append(h1, a.title);
                    if(!a.noMenuEntry) D.append(menu, D.append(D.addClass(D.span(),'current-app'), a.name));
                    const t = document.querySelector('title');
                    if(t) D.append(D.clearElement(t), a.title);
                    if(a.headerFg) DV.setColorClass(header, a.headerFg);
                    if(a.headerBg) DV.setBgClass(header, a.headerBg);
                    if(a.addStatusMessageArea) self._addStatusMessageArea()
                }else{
                    if(!a.noMenuEntry){
                        D.append(menu, D.a(a.href, a.name));
                    }
                }                
            });
            header.insertBefore(menu, header.firstChild);
            header.insertBefore(h1, header.firstChild);
            delete this._addStatusMessageArea;

            if(!thisApp) return this;
            if(!thisApp.toolbar) return this;

            const toolbar = this.e.toolbar = D.attr(
                D.fieldset("Options"), 'id', 'dv-app-toolbar'
            ), cbBorders = D.attr(
                D.checkbox(1,thisApp.cardBorders || !!body.classList.contains('card-borders')),
                'id', 'cb-card-borders'
            ), tools = D.div();
            /* Reminder to self: browsers handle multiple children
               in a toolset differently, so we wrap them in a
               div. */
            D.append(toolbar, tools);
            D.append(tools,D.append(
                D.addClass(D.span(),'input-with-label'),
                cbBorders,
                D.label(cbBorders, "Card borders?")
            ));
            const borderCheck = function(){
                D[cbBorders.checked ? 'removeClass' : 'addClass'](body, 'no-card-borders');
            }
            cbBorders.addEventListener('click', borderCheck);
            borderCheck();
            D.append(header, toolbar);
            return this;
        },

        /**
           Adds some common text to this.e.appHeader for
           card-rendering apps.
        */
        addPrintNotice: function(){
            if(!this.e.appHeader) return this;
            D.append(this.e.appHeader, D.append(
                D.p(),
                "When printing this page, this header will not be included in "+
                    "the printout. Be sure to see ",
                D.a("index.html","the main documentation"),
                " for ",
                D.a("index.html#printing", "caveats about printing"),
                "."
            ));
            return this;
        },
        
        /**
           For each card wrapper element in the DOM, this installs a
           double-click handler which will remove that element from
           the DOM when it's double-clicked. It will only add the handler
           to a given element a single time. Returns this.
        */
        installDoubleClickRemove: function(){
            const cw = document.querySelectorAll(
                '.dv-card-wrapper-mini, .dv-card-wrapper-standard'
            );
            if(!cw) return this;
            const rxWrap = /dv-card-wrapper/;
            const findWrapper = function(e){
                while(e){
                    if(rxWrap.test(e.className)) return e;
                    e = e.parentNode;
                }
            };
            const klick = function(ev){
                D.remove(findWrapper(ev.target));
            };
            let count = 0;
            cw.forEach(function(e){
                if(!e.dataset.dblClickRemove){
                    ++count;
                    e.addEventListener('dblclick', klick, true);
                    e.dataset.dblClickRemove = 1;
                }
            });
            if(count>0 && this.e.appHeader && !this.e.appHeader.dataset.dblClickRemove){
                this.e.appHeader.dataset.dblClickRemove = 1;
                const msg = D.append(
                    D.p(),
                    "Double-click individual cards to remove them from "+
                    "the page before printing it."
                );
                this.appendToHeader(msg);
            }
            return this;
        },
        
        /**
           Looks for elements matching klass.css.scriptTagSelector,
           parses their content as JSON, and replaces them with the
           klass card type rendering. If a tag contains an array of
           card objects, each one is rendered. The containing script
           elements are then removed from the DOM.

           For each script tag it creates on TABLE element, the rows
           of which contain rendered cards in the left column and
           their JSON source code in the right.

           Returns this.
        */
        scriptTagsToTables: function(klass){
            const elist = document.querySelectorAll(klass.css.scriptTagSelector);
            if(!elist) return this;
            elist.forEach(function(e){
                const json = JSON.parse(e.innerText);
                const tbl = D.addClass(D.table(),'card-script'),
                      tb = D.tbody(tbl);
                e.parentNode.insertBefore(tbl, e);
                ((json instanceof Array) ? json : [json]).forEach(function(j){
                    const tr = D.tr(tb), td1 = D.td(tr), td2 = D.td(tr);
                    const cw = D.append(klass.newWrapper(), new klass(j).e);
                    D.append(td1, cw);
                    D.append(td2, D.text(JSON.stringify(j,undefined,2)));
                });
                D.remove(e);
            });
            return this;
        },

        /**
           Looks for elements matching klass.css.scriptTagSelector,
           parses their content as JSON, and calls the given callback
           function for each card object parsed. If a tag contains an
           array of card objects, each one is passed separately to the
           callback, along with the SCRIPT DOM element as the second
           parameter (for use in knowing where to inject the card into
           the DOM, should that be the intention). The containing
           script elements are then removed from the DOM. Returns
           this.
        */
        extractScriptCards: function(klass, callback){
            const elist = document.querySelectorAll(klass.css.scriptTagSelector);
            if(!elist) return this;
            elist.forEach(function(e){
                const json = JSON.parse(e.innerText);
                ((json instanceof Array) ? json : [json]).forEach(function(j){
                    callback(j, e);
                });
                D.remove(e);
            });
            return this;
        },

        /**
           Looks for elements matching klass.css.scriptTagSelector,
           parses their content as JSON, and replaces them with the
           klass card type rendering. If a tag contains an array of
           card objects, each one is rendered. The containing script
           elements are then removed from the DOM. Returns this.

           If addToList is truthy it is assumed to be an array, and the
           parsed JSON form of each card is appended to it.
        */
        replaceScriptTags: function(klass,addToList){
            return this.extractScriptCards(klass, function(j,e){
                const cw = D.append(klass.newWrapper(), new klass(j).e);
                e.parentNode.insertBefore(cw, e);
                if(addToList) addToList.push(j);
            });
        },

        /**
           Expects to be given a card class type and a list of plain
           card objects.  Returns an array of card wrapper DOM
           elements wrapping up those (rendered) cards.
        */
        cardsToWrappers: function(klass, cardList){
            const rc = [];
            cardList.forEach(function(c){
                rc.push(D.append(klass.newWrapper(), new klass(c).e))
            });
            return rc;
        },

        /**
           Renders a list of card objects of the type specified by the
           2nd parameter (e.g. DV.SQCard or DV.HQCard) at the given
           position in the DOM, which may either be a DOM element or a
           selector for a single element. Returns this.
        */
        injectCards: function(injectPos, klass,cardList){
            injectPos = (
                'string'===typeof injectPos
                    ? document.querySelector('#js-cards')
                    : injectPos
            );
            if(injectPos){
                cardList.forEach(function(c){
                    const cw = D.append(klass.newWrapper(), new klass(c).e);
                    injectPos.parentNode.insertBefore(cw, injectPos);
                });
            }
            return this;
        },

        /**
           Helper for sidequests.html, heroicquests.html, and similar
           pages. Processes card-containing SCRIPT tags, optionally
           installs the given list of plain card objects at the
           position of #js-cards in the DOM, and installs a
           double-click-to-remove handler on all cards.

           This function sets this.scriptCards to an array of all
           plain object cards found in SCRIPT tags, in the order they
           were found.
        */
        setupFanwareCardPage: function(pageId,klass,cardList){
            this.scriptCards = [];
            this.addPrintNotice()
                .setupAppHeader(pageId)
                .replaceScriptTags(klass, this.scriptCards);
            if(cardList) this.injectCards('#js-cards', klass, cardList);
            this.installDoubleClickRemove();
        },
        
        init: function(){
            throw new Error("DV.App.init() must be overridden before calling it.");
        }
    };


    /* Mappings of deck.cardType to card class... */
    DV.SQCard.dbCardType = 1;
    DV.HQCard.dbCardType = 2;
    DV.App.config.cardTypes = {};
    [DV.SQCard, DV.HQCard].forEach((t)=>DV.App.config.cardTypes[t.dbCardType] = t);

    
})(window);

/* The DV app ajax-related bits... */
(function(DV){
    "use strict";
    /**
       fetch() is an HTTP request/response mini-framework
       similar (but not identical) to the not-quite-ubiquitous
       window.fetch().

       JS usages:

       fetch( URI [, onLoadCallback] );

       fetch( URI [, optionsObject = {}] );

       Noting that URI must be relative or the client must install a function
       fetch.urlTransform() which transforms it to fit project conventions
       (see example/default implementation below).

       The optionsObject may be an onload callback or an object with any
       of these properties:

       - onload: callback(responseData) (default = output response to the
       console). In the context of the callback, the options object is
       "this", noting that this call may have amended the options object
       with state other than what the caller provided.

       - onerror: callback(Error object) (default =
       fetch.onerror()). Triggered if the request generates any
       response other than HTTP 200 or suffers a connection error or
       timeout while awaiting a response. In the context of the
       callback, the options object is "this".

       - method: 'POST' | 'GET' (default = 'GET'). CASE SENSITIVE!

       - payload: anything acceptable by XHR2.send(ARG) (DOMString,
       Document, FormData, Blob, File, ArrayBuffer), or a plain object or
       array, either of which gets JSON.stringify()'d. If payload is set
       then the method is automatically set to 'POST'. By default XHR2
       will set the content type based on the payload type. If an
       object/array is converted to JSON, the contentType option is
       automatically set to 'application/json', and if JSON.stringify() of
       that value fails then the exception is propagated to this
       function's caller.

       - contentType: Optional request content type when POSTing. Ignored
       if the method is not 'POST'.

       - responseType: optional string. One of ("text", "arraybuffer",
       "blob", or "document") (as specified by XHR2). Default = "text".
       As an extension, it supports "json", which tells it that the
       response is expected to be text and that it should be JSON.parse()d
       before passing it on to the onload() callback. If parsing of such
       an object fails, the onload callback is not called, and the
       onerror() callback is passed the exception from the parsing error.

       - urlParams: string|object. If a string, it is assumed to be a
       URI-encoded list of params in the form "key1=val1&key2=val2...",
       with NO leading '?'.  If it is an object, all of its properties get
       converted to that form. Either way, the parameters get appended to
       the URL before submitting the request.

       - requestHeaders: an optional object mapping HTTP request
       header names to values. If Content-Type is set this way, it
       will override any value set via the contentType
       property. Ignored if the method is not 'POST'.

       - responseHeaders: If true, the onload() callback is passed an
       additional argument: a map of all of the response headers. If it's
       a string value, the 2nd argument passed to onload() is instead the
       value of that single header. If it's an array, it's treated as a
       list of headers to return, and the 2nd argument is a map of those
       header values. When a map is passed on, all of its keys are
       lower-cased. When a given header is requested and that header is
       set multiple times, their values are (per the XHR docs)
       concatenated together with ", " between them.

       - beforesend/aftersend: optional callbacks which are called
       without arguments immediately before the request is submitted
       and immediately after it is received, regardless of success or
       error. In the context of the callback, the options object is
       the "this". These can be used to, e.g., keep track of in-flight
       requests and update the UI accordingly, e.g. disabling/enabling
       DOM elements. Any exceptions thrown in an beforesend are passed
       to the onerror() handler and cause the fetch() to prematurely
       abort. Exceptions thrown in aftersend are currently silently
       ignored (feature or bug?).

       - timeout: integer in milliseconds specifying the XHR timeout
       duration. Default = fetch.timeout.

       When an options object does not provide
       onload/onerror/beforesend/aftersend handlers of its own, this
       function falls back to defaults which are member properties of
       this function with the same name, e.g. fetch.onload(). The
       default onload/onerror implementations route the data through
       the dev console. The default beforesend/aftersend are
       no-ops. Individual pages may overwrite those members to provide
       default implementations suitable for the page's use,
       e.g. keeping track of how many in-flight ajax requests are
       pending.

       Note that this routine may add properties to the 2nd argument, so
       that instance should not be kept around for later use.

       Returns void, noting that the XHR request is asynchronous,
       and still in transit (or has yet to be sent) when that happens.
    */
    const fetch = function f(uri,opt){
        if(!f.parseResponseHeaders){
            f.parseResponseHeaders = function(h){
                const rc = {};
                if(!h) return rc;
                const ar = h.trim().split(/[\r\n]+/);
                ar.forEach(function(line) {
                    const parts = line.split(': ');
                    const header = parts.shift();
                    const value = parts.join(': ');
                    rc[header.toLowerCase()] = value;
                });
                return rc;
            };
        }
        if(!opt) opt = {};
        else if('function'===typeof opt) opt={onload:opt};
        if(!opt.onload) opt.onload = f.onload;
        if(!opt.onerror) opt.onerror = f.onerror;
        if(!opt.beforesend) opt.beforesend = f.beforesend;
        if(!opt.aftersend) opt.aftersend = f.aftersend;
        let payload = opt.payload, jsonResponse = false;
        if(undefined!==payload){
            opt.method = 'POST';
            if(!(payload instanceof FormData)
               && !(payload instanceof Document)
               && !(payload instanceof Blob)
               && !(payload instanceof File)
               && !(payload instanceof ArrayBuffer)
               && ('object'===typeof payload
                   || payload instanceof Array)){
                payload = JSON.stringify(payload);
                opt.contentType = 'application/json';
            }
        }
        const url=[f.urlTransform(uri,opt.urlParams)],
              x=new XMLHttpRequest();
        if('json'===opt.responseType){
            /* 'json' is an extension to the supported XHR.responseType
               list. We use it as a flag to tell us to JSON.parse()
               the response. */
            jsonResponse = true;
            x.responseType = 'text';
        }else{
            x.responseType = opt.responseType||'text';
        }
        x.ontimeout = function(){
            try{opt.aftersend()}catch(e){/*ignore*/}
            opt.onerror(new Error("XHR timeout of "+x.timeout+"ms expired."));
        };
        x.onreadystatechange = function(){
            if(XMLHttpRequest.DONE !== x.readyState) return;
            try{opt.aftersend()}catch(e){/*ignore*/}
            if(200!==x.status){
                let err;
                try{
                    const j = JSON.parse(x.response);
                    if(j.error) err = new Error(j.error);
                    else if(j.exception && j.exception.message) err = new Error(j.exception.message);
                }catch(ex){/*ignore*/}
                opt.onerror(err || new Error("HTTP response status "+x.status+"."));
                return;
            }
            const orh = opt.responseHeaders;
            let head;
            if(true===orh){
                head = f.parseResponseHeaders(x.getAllResponseHeaders());
            }else if('string'===typeof orh){
                head = x.getResponseHeader(orh);
            }else if(orh instanceof Array){
                head = {};
                orh.forEach((s)=>{
                    if('string' === typeof s) head[s.toLowerCase()] = x.getResponseHeader(s);
                });
            }
            try{
                const args = [(jsonResponse && x.response)
                              ? JSON.parse(x.response) : x.response];
                if(head) args.push(head);
                opt.onload.apply(opt, args);
            }catch(e){
                opt.onerror(e);
            }
        };
        try{opt.beforesend()}
        catch(e){
            opt.onerror(e);
            return;
        }
        x.open(opt.method||'GET', url.join(''), true);
        if('POST'===opt.method){
            if('string'===typeof opt.contentType){
                x.setRequestHeader('Content-Type',opt.contentType);
            }
            if(opt.requestHeaders){
                let k;
                for(k in opt.requestHeaders){
                    x.setRequestHeader(k, opt.requestHeaders[k]);
                }
            }
        }
        x.timeout = +opt.timeout || f.timeout;
        if(undefined!==payload) x.send(payload);
        else x.send();
    };

    /**
       fetch.urlTransform() must refer to a function which accepts a
       URI/URL and an optional set of URL parameters to pass with it
       (in the form a of a string ("a=b&c=d...") or an object of
       key/value pairs (which it converts to such a string)), and
       returns the resulting URL or URI as a string. Depending on the
       site, it may enforce that the given URI is relative.
    */  
    fetch.urlTransform = function(u,opt){
        if(!opt) return u;
        const rc = [u];
        rc.push('?');
        if('string' === typeof opt){
            rc.push(opt);
        }else{
            let k, i = 0;
            for( k in opt ){
                if(i++) rc.push('&');
                rc.push(encodeURIComponent(k),'=',encodeURIComponent(opt[k]));
                }
        }
        return rc.join('');
    };
    fetch.beforesend = function(){};
    fetch.aftersend = function(){};
    fetch.onload = (r)=>console.debug('whutil.fetch() XHR response:',r);
    fetch.onerror = function(e/*exception or similar error object*/){
        console.error("whutil.fetch() XHR error:",e);
    };
    fetch.timeout = 15000/* Default timeout, in ms. */;

    DV.App.fetch = fetch;
})(window.DarkVenture);

(function(DV){
    /* ************************************************************ */
    /* Dark Venture-specific ajax bits...                           */
    /* ************************************************************ */
    DV.App.config.cgiScript = (function(){
        const rxDoc = /\/doc\/.+/;
        const cgiScript = 'dv-api.cgi';
        if(rxDoc){
            /* This assumes we're running via a fossil SCM instance
               and that the HTML/CSS/JS is being served via a /doc/VERSION
               link. */
            return window.location.pathname.replace(rxDoc, '')+'/ext/'+cgiScript;
        }else if("file:" !== window.location.protocol){
            /* Assume cgi-bin/. is adjacent to dv.d/. adjacent to this HTML file, for current lack of a 
               better option. */
            return window.location.pathname.replace(/[^\/]+$/,'')+
                DV.MarkerUpper.prototype.libDirUri+'../cgi-bin/'+cgiScript;
        }else{
            /* Last-ditch effort for file:// ... access, but the remote may not exist or
               it may be rejected outright by the browser because it's a CORS violation. */
            return "https://fossil.wanderinghorse.net/r/darkventure/ext/"+cgiScript;
        }
    })();

    const normalizeAjaxOpt = function(opt){
        if(!opt) opt = {};
        else if('function'===typeof opt) opt={onload:opt};
        return opt;
    }

    /**
       Ajax-specific proxy for DV.App.fetch(), documented above, which
       prepends the DV ajax API CGI path to the given path (which must
       not start with a slash).
    */
    const DAJ = DV.App.ajax = function(path,opt){
        opt = normalizeAjaxOpt(opt);
        opt.responseType = 'json';
        return DV.App.fetch(DV.App.config.cgiScript+'/'+path, opt);
    };
    const argsToArray = (a)=>Array.prototype.slice.call(a,0);
    DAJ.options = {
        /* Add any overrides to fetch.XYZ default options here. */
        timeout: 12000,
        onerror: function(){DV.App.error.apply(DV.App, argsToArray(arguments))}
    };
    Object.keys(DAJ.options).forEach(function(k){
        fetch[k] = DAJ.options[k];
    });

    /**
       A proxy for DV.App.ajax() which invokes the /decks route.
       cardType may be falsy (all card types), 1 (Side Quest), or 2
       (Heroic Quests). It returns an array of all matching deck
       objects, in an unspecified order. If includeCards is truthy
       then the cards in each deck (if any) are also fetched. The 3rd
       argument is an options object or onload() callback for ajax()'s
       2nd argument.

       Note that DV.SQCard and HQCard both have a dbCardType property
       which holds the integer value of their card type.
    */
    DAJ.getDecks = function(cardType, includeCards, opt){
        opt = normalizeAjaxOpt(opt);
        opt.urlParams = {
            cardType: cardType || 0,
            cards: !!includeCards
        };
        DAJ('decks',opt);
    };

    /**
       Submits a single card or array of cards for saving by the
       back-end. The 2nd argument is interpreted as for the ajax()
       method. Saving is atomic and either all cards are saved or none
       of them are. On success the onload() callback, if any, is
       passed either the newly-saved card object (which will have been
       normalized and modified slightly by the server) or, if passed
       an array, an array of such objects.

       Note that cards MUST have a 'meta' property (object) with a
       'deck' property set before they can be saved. Without that, the
       server will reject them. If card.meta has no 'uid' value then
       it is considered to be a new card and a uid will be assigned by
       the server. If card.meta has a uid value it may refer to an
       existing card (which it will overwrite, if able) or it may be a
       new card.
    */
    DAJ.saveCard = function(cards, opt){
        opt = normalizeAjaxOpt(opt);
        opt.payload = cards;
        DAJ('card', opt);
    };

    /** Only for use in testing and populating the database. */
    DV.App._batchSaveCards = function(cardList, opt){
        cardList.forEach(function(c){
            if(!c.meta) c.meta = {};
            if(!c.meta.deck) c.meta.deck = opt.deck;
        });
        DAJ.saveCard(cardList, function(r){
            DV.App.message("Saved",r.length,"card(s).");
        });
    };

})(window.DarkVenture);
