-- These logging parts of the schema are attached to a different
-- db file, to keep the core db size/backups from exploding.
-- This info doesn't strictly need to be backed up, but has proven
-- useful for recovery shortly after "mistakes were made."

-- ------------------------------------------------------------------------
-- The card log (clog) table is used to record saves and deletions,
-- noting that a log for saving records the new contents, not the old.
-- Deletion is logged with a card's pre-deletion content.
DROP TABLE IF EXISTS clog;
CREATE TABLE clog (
  id INTEGER primary KEY autoincrement DEFAULT NULL,
  parent INTEGER DEFAULT NULL REFERENCES clog(id) ON UPDATE CASCADE ON DELETE CASCADE, -- parent version
  mtime INTEGER NOT NULL, -- Unix Epoch time w/ seconds precision
  op TEXT,  -- logged operation name. e.g. "save", "update", "delete", "snapshot"
  deck TEXT COLLATE NOCASE, -- qcard.deck value
  uid TEXT, -- qcard.uid value
  json TEXT -- full JSON of card body
);

DROP INDEX IF EXISTS x_clog_deck_uid;
CREATE INDEX x_clog_deck_uid ON clog(deck,uid);

DROP VIEW IF EXISTS vclog;
CREATE VIEW vclog AS
SELECT id as id,
  parent as parent,
  strftime('%Y-%m-%d %H:%M:%S',mtime,'unixepoch','utc') as mtime,
  op as op,
  deck as deck,
  uid as uid
FROM clog ORDER BY mtime desc, id DESC;

-- Extended vclog which includes child-parent mapping and orders the
-- results such that changes to a given card are grouped together.
DROP VIEW IF EXISTS vclogh;
CREATE VIEW vclogh AS
with
cbranch(id,parent,mtime,op,deck,uid) as (
  SELECT id,NULL, strftime('%Y-%m-%d %H:%M:%S',mtime,'unixepoch','utc'),
    op, deck, uid FROM clog WHERE parent IS NULL
  UNION ALL
  SELECT cl.id, cl.parent, strftime('%Y-%m-%d %H:%M:%S',cl.mtime,'unixepoch','utc'),
  cl.op, cl.deck, cl.uid
  FROM clog cl, cbranch cb WHERE cl.parent=cb.id
)
SELECT mtime, deck, uid, op, id, parent
FROM cbranch
ORDER BY deck, uid, mtime DESC, id DESC
;
